#include "stdafx.h"

#include <string>

#include "ProfileCreator.h"

  // Constructor

  ProfileCreator::ProfileCreator() {

   ZERO_NEFF = 0.0;
   MAX_SIM = 18.0;
   SIM_STEP = 0.5;
   SIM_INIT = 2;
   MAX_NEFF = 100.0;
   pseudocount = 5;
   use_sqrt_neff = true;
   lambda = 0.3f;

   alignment_length = 0;

   alphabet = NULL;
   alignment = NULL;
   subst_matrix = NULL;

   // Init data structures

   NEFF = new float[1];

   similarities = new float*[1];
   distances = new float*[1];
   pseudofreqs = new float*[1];
   freqs = new float*[1];
   n_eff = new float*[1];
   probs = new float*[1];
   profile_matrix = new float*[1];

   for (int posIndex = 0; posIndex<1; posIndex++) {
    similarities[posIndex] = new float[1];
    distances[posIndex] = new float[1];
    pseudofreqs[posIndex] = new float[1];
    freqs[posIndex] = new float[1];
    n_eff[posIndex] = new float[1];
    probs[posIndex] = new float[1];
    profile_matrix[posIndex] = new float[1];
   }

  }

  // Destructor

  ProfileCreator::~ProfileCreator() {

   int posIndex;

   // Free the data structures

   for (posIndex = 0; posIndex<alignment_length; posIndex++) {
    delete[] similarities[posIndex];
    delete[] distances[posIndex];
    delete[] pseudofreqs[posIndex];
    delete[] freqs[posIndex];
    delete[] n_eff[posIndex];
    delete[] probs[posIndex];
   }

   delete[] similarities;
   delete[] distances;
   delete[] pseudofreqs;
   delete[] freqs;
   delete[] n_eff;
   delete[] probs;
   delete[] NEFF;


   // Free the subst matrix

   for (posIndex = 0; posIndex<subst_matrix_size; posIndex++) {
    delete[] subst_matrix[posIndex];
   }

   delete[] subst_matrix;

   // Free the alignment

   for (posIndex = 0; posIndex<alignment_size; posIndex++) {
    delete[] alignment[posIndex];
   }

   delete[] alignment;

  }

  //Methods

  int ProfileCreator::getAlphabetSymbolIndex(char symbol) {

   for (int symbolIndex = 0; symbolIndex<alphabet_length; symbolIndex++) {
    if (symbol == alphabet[symbolIndex]) {
     return symbolIndex;
    }
   }

   return -1;
  }

  float ProfileCreator::getAlphabetSymbolFreq(int symbolIndex) {

   if (symbolIndex != -1) {
    return symbol_freqs[symbolIndex];
   }
   else {
    return 0.0;
   }
  }

  bool ProfileCreator::loadTableNEff(const char* fileName) {

   ifstream table_file(fileName, ios::in);

   if (table_file) {
    int maxIndex = (int) (MAX_SIM/SIM_STEP) - 2;
    for (int index = 0; index<maxIndex; index++) {
     table_file >> table_neff[index];
    }
    table_file.close();

    return true;
   }

   return false;
  }

  bool ProfileCreator::loadAlphabet(const char* fileName) {

   // Open alphabet file

   ifstream alphabet_file(fileName,ios::in);

   if (!alphabet_file) {
    return false;
   }

   // Free the memory - if need

   if (alphabet) {
    delete[] alphabet;
    delete[] symbol_freqs;
   }

   int len = 1000;
   char* tmp_str = new char[len];

   // Read alphabet length

   alphabet_file.getline(tmp_str,len);
   alphabet_length = atoi(tmp_str);

   // Allocate the memory

   alphabet = new char[alphabet_length];
   symbol_freqs = new float[alphabet_length];

   // Read symbols & frequencies

   alphabet_file.getline(alphabet,len);

   int symbolIndex;
   float value;

   for (symbolIndex = 0; symbolIndex<alphabet_length; symbolIndex++) {
    alphabet_file.getline(tmp_str,len);
    value = (float) atof(tmp_str);
    symbol_freqs[symbolIndex] = value;
   }

   alphabet_file.close();

   delete[] tmp_str;

   return true;
  }

  bool ProfileCreator::loadAlignment(const char* fileName) {

   int posIndex;

   // Open alignment file

   ifstream alignment_file(fileName, ios::in);

   if (!alignment_file) {
    return false;
   }

   // Free the memory for alignment - if need

   if (alignment) {
    for (posIndex = 0;posIndex<alignment_size; posIndex++) {
     delete[] alignment[posIndex];
    }
    delete[] alignment;
   }

   // Free the memory of data_structures

   for (posIndex = 0; posIndex<alignment_length; posIndex++) {
    delete[] similarities[posIndex];
    delete[] distances[posIndex];
    delete[] pseudofreqs[posIndex];
    delete[] freqs[posIndex];
    delete[] n_eff[posIndex];
    delete[] probs[posIndex];
    delete[] profile_matrix[posIndex];
   }

   delete[] similarities;
   delete[] distances;
   delete[] pseudofreqs;
   delete[] freqs;
   delete[] n_eff;
   delete[] probs;
   delete[] profile_matrix;

   delete[] NEFF;

   int len = 1000;
   char* tmp_str = new char[len];

   // Read alignment parameters

   alignment_file.getline(tmp_str,len);
   alignment_size = atoi(tmp_str);
   alignment_file.getline(tmp_str,len);
   alignment_length = atoi(tmp_str);

   // Allocate the memory for alignment

   alignment = new char*[alignment_size];

   for (posIndex = 0; posIndex<alignment_size; posIndex++) {
    alignment[posIndex] = new char[alignment_length];
   }

   // Read sequences

   for (posIndex = 0; posIndex<alignment_size; posIndex++) {
    alignment_file.getline(tmp_str,len);
    strncpy(alignment[posIndex],tmp_str,alignment_length);
   }

   alignment_file.close();

   delete[] tmp_str;

   // Allocate the memory for data_structures

   NEFF = new float[alignment_length];

   similarities = new float*[alignment_length];
   distances = new float*[alignment_length];
   pseudofreqs = new float*[alignment_length];
   freqs = new float*[alignment_length];
   n_eff = new float*[alignment_length];
   probs = new float*[alignment_length];
   profile_matrix = new float*[alignment_length];

   for (posIndex = 0; posIndex<alignment_length; posIndex++) {
    similarities[posIndex] = new float[alphabet_length];
    distances[posIndex] = new float[alphabet_length];
    pseudofreqs[posIndex] = new float[alphabet_length];
    freqs[posIndex] = new float[alphabet_length];
    n_eff[posIndex] = new float[alphabet_length];
    probs[posIndex] = new float[alphabet_length];
    profile_matrix[posIndex] = new float[alphabet_length];
   }

   return true;
  }

  bool ProfileCreator::loadSubstMatrix(const char* fileName) {

   // Open subst_matrix file

   ifstream matrix_file(fileName,ios::in);

   if (!matrix_file) {
    return false;
   }

   // Free the memory - if need

   int posIndex;

   if (subst_matrix) {
    for (posIndex = 0; posIndex<subst_matrix_size; posIndex++) {
     delete[] subst_matrix[posIndex];
    }
    delete[] subst_matrix;
   }

   // Read subst matrix size

   int len = 1000;
   char* tmp_str = new char[len];

   matrix_file.getline(tmp_str,len);
   subst_matrix_size = atoi(tmp_str);

   delete[] tmp_str;

   // Allocate the memory

   subst_matrix = new float*[subst_matrix_size];
   for (posIndex = 0; posIndex<subst_matrix_size; posIndex++) {
    subst_matrix[posIndex] = new float[subst_matrix_size];
   }

   // Read subst matrix

   for (int posIndex1 = 0; posIndex1<subst_matrix_size; posIndex1++) {
    for (int posIndex2 = 0; posIndex2<subst_matrix_size; posIndex2++) {
     matrix_file >> subst_matrix[posIndex1][posIndex2];
    }
   }

   matrix_file.close();

   return true;
  }

  void ProfileCreator::setPseudocount(int newPseudocount) {

   if (newPseudocount>0) {
    pseudocount = newPseudocount;
   }
   else {
    pseudocount = 1;
   }
  }

  void ProfileCreator::setLambda(float newLambda) {

   lambda = newLambda;
  }

  void ProfileCreator::setSqrtNeffFlag(bool newFlag) {

   use_sqrt_neff = newFlag;
  }

  float ProfileCreator::getTableNEff(float similarity) {

   if (similarity<1.0) {
    if (similarity == 0.0) {
     return ZERO_NEFF;
    }
    else {
     return 1.0;
    }
   }
   else {
    if (similarity<MAX_SIM) {
     double work_value = (similarity/SIM_STEP)-SIM_INIT;
     double int_part;
     int work_index = (int) (floor(2*modf(work_value,&int_part))+floor(int_part));
     return table_neff[work_index];
    }
    else {
     return MAX_NEFF;
    }
   }
  }

  void ProfileCreator::calculateSimilarities() {

   int posIndex;

   // Prepare aln_matrix

   int** aln_matrix = new int*[alignment_length];

   for (posIndex = 0; posIndex<alignment_length; posIndex++) {
    aln_matrix[posIndex] = new int[alignment_size];
   }

   // Fill aln_matrix

   for (posIndex = 0; posIndex<alignment_length; posIndex++) {
    for (int seqIndex = 0; seqIndex<alignment_size; seqIndex++) {
     aln_matrix[posIndex][seqIndex] = getAlphabetSymbolIndex(alignment[seqIndex][posIndex]);
    }
   }

   int* seqs_indicies = new int[alignment_size];
   int seqs_count;
   float symbols_num;
   int length;
   int index;
   int symbolIndex;
   int symbols_count;

   for (int pos_index = 0; pos_index<alignment_length; pos_index++) {
    for (int res_index = 0; res_index<alphabet_length; res_index++) {

     // Prepare set of seqs with residue #res_index at position #pos_index
     seqs_count = 0;
     for (int seq_index = 0; seq_index<alignment_size; seq_index++) {
      if (getAlphabetSymbolIndex(alignment[seq_index][pos_index]) == res_index) {
       seqs_indicies[seqs_count] = seq_index;
       seqs_count++;
      }
     }

     // Calc seqs_similarity
     if (seqs_count == 0) {
      similarities[pos_index][res_index] = 0.0;
     }
     else
     if (seqs_count == 1) {
      similarities[pos_index][res_index] = 1.0;
     }
     else {

      symbols_num = 0;
      length = 0;

      for (int posIndex = 0; posIndex<alignment_length; posIndex++) {

       // Calc num of different acids at pos #posIndex

       symbols_count = 0;
       for (symbolIndex = 0; symbolIndex<alphabet_length; symbolIndex++) {
        index = 0;
        while (index<seqs_count) {
         if (symbolIndex == aln_matrix[posIndex][seqs_indicies[index]]) {
          symbols_count++;
          index = seqs_count;
         }
         index++;
        }
       }
       if (symbols_count>0) {
        symbols_num += symbols_count;
        length++;
       }
      }

      // Similarity = average num of different acids at positions

      if (length>0) {
       similarities[pos_index][res_index] = symbols_num/length;
      }
      else {
       similarities[pos_index][res_index] = 0.0;
      }
     }

    }
   }

   // Free the memory

   delete[] seqs_indicies;

   for (posIndex = 0; posIndex<alignment_length; posIndex++) {
    delete[] aln_matrix[posIndex];
   }

   delete[] aln_matrix;
  }

  void ProfileCreator::calculateNEff() {

   calculateSimilarities();

   int posIndex;

   for (posIndex = 0; posIndex<alignment_length; posIndex++) {
    NEFF[posIndex] = 0.0;
    for (int resIndex = 0; resIndex<alphabet_length; resIndex++) {
     n_eff[posIndex][resIndex] = getTableNEff(similarities[posIndex][resIndex]);
     NEFF[posIndex] += n_eff[posIndex][resIndex];
    }
   }

   for (posIndex = 0; posIndex<alignment_length; posIndex++) {
    if (NEFF[posIndex] != 0.0) {
     for (int resIndex = 0; resIndex<alphabet_length; resIndex++) {
      freqs[posIndex][resIndex] = n_eff[posIndex][resIndex]/NEFF[posIndex];
     }
    }
    else {
     for (int resIndex = 0; resIndex<alphabet_length; resIndex++) {
      freqs[posIndex][resIndex] = getAlphabetSymbolFreq(resIndex);
     }
    }
   }
  }

  void ProfileCreator::calculateFreqs() {

   calculateNEff();

   for (int resIndex = 0; resIndex<alphabet_length; resIndex++) {
    for (int posIndex = 0; posIndex<alignment_length; posIndex++) {
     pseudofreqs[posIndex][resIndex] = 0.0;
     for (int symbolIndex = 0; symbolIndex<alphabet_length; symbolIndex++) {
      pseudofreqs[posIndex][resIndex] += (float) ((freqs[posIndex][symbolIndex])*(exp(lambda*subst_matrix[resIndex][symbolIndex])));
     }
     pseudofreqs[posIndex][resIndex] = (pseudofreqs[posIndex][resIndex])*(getAlphabetSymbolFreq(resIndex));
    }
   }
  }

  void ProfileCreator::calculateProbs() {

   calculateFreqs();

   if (use_sqrt_neff) {
    float new_pseudocount;
    for (int resIndex = 0; resIndex<alphabet_length; resIndex++) {
     for (int posIndex = 0; posIndex<alignment_length; posIndex++) {
      new_pseudocount = (float) (pseudocount*sqrt(NEFF[posIndex]));
      probs[posIndex][resIndex] = (float) (((NEFF[posIndex]-1)*freqs[posIndex][resIndex]+new_pseudocount*pseudofreqs[posIndex][resIndex])/(NEFF[posIndex]-1.0+new_pseudocount));
     }
    }
   }
   else {
    for (int resIndex = 0; resIndex<alphabet_length; resIndex++) {
     for (int posIndex = 0; posIndex<alignment_length; posIndex++) {
      probs[posIndex][resIndex] = (float) (((NEFF[posIndex]-1)*freqs[posIndex][resIndex]+pseudocount*pseudofreqs[posIndex][resIndex])/(NEFF[posIndex]-1.0+pseudocount));
     }
    }
   }
  }

  void ProfileCreator::calculateProfile() {

   calculateProbs();

   for (int resIndex = 0; resIndex<alphabet_length; resIndex++) {
    for (int posIndex = 0; posIndex<alignment_length; posIndex++) {
     profile_matrix[posIndex][resIndex] = (float) (log(probs[posIndex][resIndex]/(getAlphabetSymbolFreq(resIndex))));
    }
   }
  }

  bool ProfileCreator::saveProfile(const char* fileName) {

   ofstream profile_file(fileName,ios::out);

   if (!profile_file) {
    return false;
   }

   int posIndex;
   int symbolIndex;

   profile_file << "\t";
   for (symbolIndex = 0; symbolIndex<alphabet_length; symbolIndex++) {
    profile_file << alphabet[symbolIndex] << "\t";
   }
   profile_file << "\n";

   for (posIndex = 0; posIndex<alignment_length; posIndex++) {

    profile_file << posIndex << "\t";

    for (symbolIndex = 0; symbolIndex<alphabet_length; symbolIndex++) {
     profile_file << profile_matrix[posIndex][symbolIndex] << "\t";
    }

    profile_file << "\n";
   }

   profile_file << "\n";

   profile_file.close();

   return true;
  }


