#ifndef ProfileCreator_H
#define ProfileCreator_H

#include <fstream.h>
#include <math.h>

class ProfileCreator {

  // Fields

  private: char* alphabet;
  private: float* symbol_freqs;
  private: int alphabet_length;

  private: char** alignment;
  private: int alignment_size;
  private: int alignment_length;

  private: float** profile_matrix;

  private: float** subst_matrix;
  private: int subst_matrix_size;

  private: bool use_sqrt_neff;
  private: float ZERO_NEFF;
  private: float MAX_SIM;
  private: float SIM_STEP;
  private: float SIM_INIT;
  private: float MAX_NEFF;
  private: static char* TableNEff_fileName;

  private: int pseudocount;
  private: float lambda;  
  private: float table_neff[100];
  public: float* NEFF;

  public: float** probs;
  public: float** n_eff;
  public: float** freqs;
  public: float** pseudofreqs;
  public: float** distances;
  public: float** similarities;

  // Constructor

  public: ProfileCreator();

  // Destructor

  public: ~ProfileCreator();

  // Methods

  private: int getAlphabetSymbolIndex(char symbol);
  private: float getAlphabetSymbolFreq(int symbolIndex);

  public: bool loadSubstMatrix(const char* fileName);
  public: bool loadAlphabet(const char* fileName);
  public: bool loadAlignment(const char* fileName);
  public: bool loadTableNEff(const char* fileName);

  public: void setPseudocount(int newPseudocount);
  public: void setLambda(float newLambda);
  public: void setSqrtNeffFlag(bool newFlag);

  public: float getTableNEff(float similarity);

  private: void calculateSimilarities();
  private: void calculateNEff();
  private: void calculateFreqs();
  private: void calculateProbs();
  public: void calculateProfile();

  public: bool saveProfile(const char* fileName);

};

#endif
